package org.bitbucket.bradleysmithllc.etlunit.maven;

import com.sun.codemodel.*;
import org.apache.axis.utils.JavaUtils;
import org.apache.commons.lang.ClassUtils;
import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.project.MavenProject;
import org.bitbucket.bradleysmithllc.etlunit.*;
import org.bitbucket.bradleysmithllc.etlunit.context.VariableContext;
import org.bitbucket.bradleysmithllc.etlunit.feature.*;
import org.bitbucket.bradleysmithllc.etlunit.io.FileBuilder;
import org.bitbucket.bradleysmithllc.etlunit.maven.util.JavaProperty;
import org.bitbucket.bradleysmithllc.etlunit.maven.util.JavaSource;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestOperation;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestValueObject;
import org.bitbucket.bradleysmithllc.etlunit.util.StringUtils;
import org.bitbucket.bradleysmithllc.etlunit.util.maven.regexp.OperationNameExpression;
import org.bitbucket.bradleysmithllc.json.validator.*;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 * Generates stubs based on json descriptors.
 *
 * @goal generate-gson-proxies
 * @phase generate-sources
 * @requiresDependencyResolution compile
 */
public class FeatureCompilerMojo
		extends AbstractMojo
{
	/**
	 * @parameter expression="${project}" default-value="${project}"
	 */
	protected MavenProject mavenProject;

	/**
	 * Location of the file.
	 *
	 * @parameter expression="${basedir}/src/main/resources"
	 * @required
	 */
	private File inputDirectory;

	private final SchemaResolver resolver = new CachingSchemaResolver();

	/**
	 * Location of the file.
	 *
	 * @parameter expression="${outputDirectory}" default-value="${project.build.directory}/generated-sources/etlunit"
	 * @required
	 */
	private File outputDirectory;

	private final List<JsonSchemaValidationException> schemaErrors = new ArrayList<JsonSchemaValidationException>();

	public void execute()
			throws MojoExecutionException
	{
		File f = outputDirectory;

		if (!f.exists())
		{
			f.mkdirs();
		}

		JCodeModel cm = new JCodeModel();

		//for (Object res : mavenProject.getResources())
		//{
		//	Resource re = (Resource) res;

		//	File fi = new File(re.getDirectory());

		if (inputDirectory.exists())
		{
			File
					services =
					new FileBuilder(inputDirectory).subdir("META-INF")
							.subdir("services")
							.name("org.bitbucket.bradleysmithllc.etlunit.feature.Feature")
							.file();

			// check for a services entry for an etlunit feature
			if (services.exists())
			{
				getLog().info("Processing services file: " + services);

				// scan this file for features
				try
				{
					BufferedReader br = new BufferedReader(new FileReader(services));

					String fname;

					while ((fname = br.readLine()) != null)
					{
						processFeatureClassName(fname, cm);
					}
				}
				catch (Exception e)
				{
					throw new MojoExecutionException("Failure reading services file", e);
				}
			}
		}
		//}

		if (schemaErrors.size() != 0)
		{
			throw new MojoExecutionException("Errors processing schema files: " + schemaErrors);
		}

		try
		{
			cm.build(outputDirectory);
		}
		catch (IOException e)
		{
			throw new MojoExecutionException("", e);
		}

		if (mavenProject != null)
		{
			mavenProject.addCompileSourceRoot(outputDirectory.getAbsolutePath());
		}
	}

	private void processFeatureClassName(String fname, JCodeModel cm)
			throws
			InstantiationException,
			IllegalAccessException,
			IOException,
			JClassAlreadyExistsException,
			JsonSchemaValidationException
	{
		// load this class into the vm
		try
		{
			Class fcl = Class.forName(fname);
			Package aPackage = fcl.getPackage();

			JPackage pack = cm.rootPackage();

			if (aPackage != null)
			{
				pack = cm._package(aPackage.getName());
			}

			processFeatureClass(fcl, pack);
		}
		catch (ClassNotFoundException cfne)
		{
			// didn't work - try to use a resource meta info instead
			URL[] urls = new URL[]{inputDirectory.toURL()};

			URLClassLoader ucl = new URLClassLoader(urls, Thread.currentThread().getContextClassLoader());

			ResourceFeatureMetaInfo rfmi = new ResourceFeatureMetaInfo(fname, ucl);

			JPackage pack = cm.rootPackage();

			String npack = ClassUtils.getPackageName(fname);

			if (npack != null && !npack.equals(""))
			{
				pack = cm._package(npack);
			}

			processMetaInfo(rfmi, pack);
		}
	}

	private void processFeatureClass(Class fcl, JPackage pack)
			throws
			IllegalAccessException,
			InstantiationException,
			IOException,
			JClassAlreadyExistsException,
			JsonSchemaValidationException
	{
		getLog().info("Generating stubs for feature: " + fcl.getSimpleName());

		Feature f = (Feature) fcl.newInstance();

		FeatureMetaInfo metaInfo = f.getMetaInfo();
		if (metaInfo != null)
		{
			processMetaInfo(metaInfo, pack);
		}
	}

	private void processMetaInfo(FeatureMetaInfo metaInfo, JPackage pack)
			throws IOException, JClassAlreadyExistsException, JsonSchemaValidationException
	{
		getLog().info("Processing meta info");

		JsonSchema vald = metaInfo.getFeatureConfigurationValidator();
		JsonValidator jv = new JsonValidator(vald, new ClasspathSchemaResolver(new URLClassLoader(new URL[]{inputDirectory.toURL()})));

		if (vald != null)
		{
			processFeatureConfiguration(metaInfo, jv.resolveSchemaReference(vald.getSchemaNode()), pack);
		}

		Map<String, FeatureOperation> exportedOperations = metaInfo.getExportedOperations();
		if (exportedOperations != null)
		{
			processOperations(metaInfo, exportedOperations, pack);
		}
	}

	private void processFeatureConfiguration(FeatureMetaInfo metaInfo, JsonSchemaObjectNode vald, JPackage pack)
			throws IOException, JClassAlreadyExistsException
	{
		getLog().info("Processing configuration");

		generateSourceFromSchema(vald,
				ClassUtils.getShortClassName(metaInfo.getDescribingClassName()) + "Configuration",
				pack);
	}

	private void processOperations(FeatureMetaInfo metaInfo, Map<String, FeatureOperation> exportedOperations, JPackage pack)
			throws IOException, JClassAlreadyExistsException, JsonSchemaValidationException
	{
		getLog().info("Processing operations");

		JPackage spack = pack.subPackage(sanitize(StringUtils.makeUnderscoreName(metaInfo.getFeatureName())));

		for (Map.Entry<String, FeatureOperation> entry : exportedOperations.entrySet())
		{
			processOperation(entry.getValue(), spack);
		}
	}

	private void processOperation(FeatureOperation value, JPackage pack)
			throws IOException, JClassAlreadyExistsException, JsonSchemaValidationException
	{
		getLog().info("Processing operation: '" + value.getName() + "'");

		if (
				value.getSignatures().size() > 0)
		{
			JDefinedClass jreqc = null;
			JDefinedClass jhandc = null;

			if (
					value.getSignatures().size() > 1 ||
							value.getSignatures().get(0).isSubOperation()
					)
			{
				String s = StringUtils.makeProperPropertyName(value.getName());
				jreqc = generateSourceFromSchema(new JsonSchema(ResourceFeatureOperation.NULL_NODE).getSchemaNode(), s + "Request", pack);
				jhandc = generateInterfaceForOperation(value.getName(), value.getName() + "Handler", pack, jreqc, true);
			}

			for (FeatureOperationSignature sig : value.getSignatures())
			{
				processOperationSignature(value, jreqc, jhandc, sig, pack);
			}
		}
	}

	private void processOperationSignature(FeatureOperation value, JDefinedClass jreqc, JDefinedClass jhandc, FeatureOperationSignature sig, JPackage pack)
			throws IOException, JClassAlreadyExistsException, JsonSchemaValidationException
	{
		getLog().info("Processing operation signature: '" + (sig.isSubOperation()
				? (value.getName() + "." + sig.getId())
				: sig.getId()) + "'");

		JsonSchema validator = sig.getValidator();
		JsonValidator jv = new JsonValidator(validator, new ClasspathSchemaResolver(new URLClassLoader(new URL[]{inputDirectory.toURL()})));

		if (validator != null)
		{
			String s = StringUtils.makeProperPropertyName(sig.isSubOperation() ? sig.getId() : (value.getName() + "Request"));
			JDefinedClass jc = generateSourceFromSchema(jv.resolveSchemaReference(validator.getSchemaNode()), s, sig.isSubOperation() ? jreqc : pack);
			generateInterfaceForOperation(sig.isSubOperation() ? sig.getId() : value.getName(), sig.isSubOperation() ? sig.getId() : (value.getName() + "Handler"), sig.isSubOperation() ? jhandc : pack, jc, false);
		}
	}

	private JDefinedClass generateInterfaceForOperation(String operationName, String className, JClassContainer cm, JDefinedClass opClass, boolean omitMethod)
			throws IOException, JClassAlreadyExistsException
	{
		className = StringUtils.makeProperPropertyName(className);

		getLog().info("Creating processor interface: " + operationName);
		JDefinedClass definedClass = cm._interface(className);
		JClass refOP = cm.owner().ref(OperationProcessor.class);
		JClass refCR = cm.owner().ref(ClassResponder.class);
		JClass refac = cm.owner().ref(ClassResponder.action_code.class);
		definedClass._extends(refOP);
		definedClass._extends(refCR);

		if (!omitMethod)
		{
			JMethod meth = definedClass.method(JMod.PUBLIC | JMod.ABSTRACT, refac, sanitize(StringUtils.makePropertyName(
					operationName)));
			meth.param(opClass, "request");
			meth.param(ETLTestOperation.class, "testOperation");
			meth.param(ETLTestValueObject.class, "valueObject");
			meth.param(VariableContext.class, "variableContext");
			meth.param(ExecutionContext.class, "executionContext");

			meth._throws(TestAssertionFailure.class);
			meth._throws(TestExecutionError.class);
			meth._throws(TestWarning.class);
		}

		return definedClass;
	}

	private String sanitize(String s)
	{
		if (JavaUtils.isJavaKeyword(s))
		{
			return "_" + s;
		}

		return s;
	}

	private String getOperationName(String name)
	{
		OperationNameExpression one = new OperationNameExpression(name);

		if (one.hasNext())
		{
			return one.getOperationName();
		}

		throw new IllegalArgumentException("Invalid schema file name: " + name);
	}

	private JDefinedClass generateSourceFromSchema(JsonSchemaObjectNode schema, String className, JClassContainer cm)
			throws IOException, JClassAlreadyExistsException
	{
		getLog().info("Creating schema container: " + className);

		Map<String, JsonSchemaObjectNode> properties = schema.getPropertySchemas();

		JavaSource source;

		if (schema.getId() != null)
		{
			source = new JavaSource(className, schema.getId());
		}
		else
		{
			source = new JavaSource(className);
		}

		for (Map.Entry<String, JsonSchemaObjectNode> entry : properties.entrySet())
		{
			String key = entry.getKey();
			if (key.startsWith("_"))
			{
				continue;
			}

			JsonSchemaObjectNode propertySchema = entry.getValue();

			JavaProperty
					property =
					new JavaProperty(key,
							propertySchema.getType().get(0),
							propertySchema.getEnumValues(),
							propertySchema.isRequired());

			source.addProperty(property);
		}

		return source.toSource(cm);
	}

	public void setOutputDirectory(File outputDirectory)
	{
		this.outputDirectory = outputDirectory;
	}

	public void setInputDirectory(File inputDirectory)
	{
		this.inputDirectory = inputDirectory;
	}
}